@@ -1,17 +1,22 @@
 import os
 import json
 import time
-from typing import Optional, Dict, Any, List
+import re
+from typing import Optional, Dict, Any, List, Union
 import httpx
 from dotenv import load_dotenv
-from apidance.models import Tweet, User
+from .models import Tweet, User
 from .exceptions import (
-    AuthenticationError,
+    TwitterPlatformError,
+    InvalidInputError,
     RateLimitError,
     InsufficientCreditsError,
     TimeoutError,
-    AccountSuspendedError,
+    ApidancePlatformError,
+    ApiKeyError,
+    AuthTokenError,
 )
+from .utils import parse_markdown_to_richtext
 
 load_dotenv()
 
@@ -40,19 +45,16 @@ class TwitterClient:
         """
         self.api_key = api_key or os.getenv("APIDANCE_API_KEY")
         if not self.api_key:
-            raise ValueError(
+            raise ApiKeyError(
                 "API key must be provided either through constructor or APIDANCE_API_KEY environment variable"
             )
 
         # Check balance
         balance = self.check_balance()
         if int(balance) < 100:
-            print(
+            raise InsufficientCreditsError(
                 f"Warning: Your API balance is low ({balance}). Please recharge your account."
             )
-            response = input("Do you want to continue? [y/N]: ")
-            if response.lower() != "y":
-                raise SystemExit("Operation cancelled by user.")
 
         self.base_url = base_url
         self.client = httpx.Client()
@@ -101,61 +103,54 @@ class TwitterClient:
         if attempt >= self.max_retries:
             return False
 
+        # Try to parse the response as JSON
+        response_data = None
         try:
             response_data = response.json()
-
-            # Handle Twitter API style errors
-            if "errors" in response_data:
-                error = response_data["errors"][0]
-                if error.get("code") == 88:
-                    if attempt == self.max_retries:
-                        raise RateLimitError(
-                            "Rate limit exceeded. Please try again later."
-                        )
-                    return True
-                elif error.get("code") == 32:
-                    raise AuthenticationError(
-                        "Could not authenticate you. Please check your X_AUTH_TOKEN."
-                    )
-                elif error.get("code") == 366:
-                    print("NumericString value expected.")
-                    return False
-                elif error.get("code") == 64:
-                    raise AccountSuspendedError(
-                        "Your account is suspended and is not permitted to access this feature."
-                    )
-                elif error.get("code") == 139:
-                    return
-
-            # Handle Apidance API style errors
-            if isinstance(response_data, dict):
-                if response_data.get("code") == 401:
-                    if (
-                        "insufficient api counts"
-                        in response_data.get("msg", "").lower()
-                    ):
-                        raise InsufficientCreditsError(
-                            "Insufficient API credits. Please contact support via Telegram: @shingle"
-                        )
-
-            # Handle other error cases
-            if (
-                response.text == "local_rate_limited"
-                or response.text == "null"
-                or response_data is None
-            ):
-                return True
-
-            # If response is normal, no need to retry
-            if response_data and not response_data.get("errors"):
-                return False
-
-            return True
-
         except json.JSONDecodeError:
             # JSON parse error might be temporary, allow retry
             return True
 
+        # Handle Twitter-style errors (with "errors" array)
+        if response_data is not None and "errors" in response_data:
+            error = response_data["errors"][0]
+            error_code = error.get("code")
+
+            # Rate limit error
+            if error_code == 88:
+                if attempt == self.max_retries:
+                    raise RateLimitError("Rate limit exceeded. Please try again later.")
+                return True
+            # Other Twitter platform errors
+            elif error_code == 366:
+                raise InvalidInputError(error.get("message", ""))
+            elif error_code == 139:  # Tweet already favorited
+                return False
+            elif (
+                error_code == 64
+            ):  # Account is suspended and is not permitted to access this feature, retry
+                return True
+            else:
+                raise TwitterPlatformError(error.get("message", ""))
+
+        # Handle Apidance API style errors
+        if response_data is not None and response_data.get("code", 0) > 0:
+            msg = response_data.get("msg", "").lower()
+            if "insufficient api counts" in msg:
+                raise InsufficientCreditsError("Insufficient api credits")
+            else:
+                raise ApidancePlatformError(msg)
+
+        # Handle local rate limiting
+        if response.text == "local_rate_limited":
+            return True
+
+        # Handle specific error cases based on response data
+        if "data" in response_data:
+            return False
+
+        return False
+
     def _make_request(self, method: str, endpoint: str, **kwargs) -> Dict[str, Any]:
         """Make an API request with retry logic.
 
@@ -168,10 +163,12 @@ class TwitterClient:
             API response data
 
         Raises:
-            AuthenticationError: When authentication fails
-            TimeoutError: When request times out
+            AuthTokenError: When authentication token is invalid
+            PremiumRequiredError: When Premium+ is required
             RateLimitError: When rate limit is exceeded
             InsufficientCreditsError: When API credits are depleted
+            TimeoutError: When request times out
+            ApidancePlatformError: When the platform returns other errors
         """
         url = f"{self.base_url}{endpoint}"
 
@@ -186,16 +183,31 @@ class TwitterClient:
         auth_required_endpoints = [
             "/graphql/FollowersYouKnow",
             "/graphql/CreateTweet",
+            "/graphql/CreateNoteTweet",
             "/graphql/FavoriteTweet",
         ]
         if endpoint in auth_required_endpoints:
             if token:
+                if not re.match(r"^[0-9a-f]{40}$", token):
+                    raise AuthTokenError(
+                        f"Invalid AuthToken format. Expected 40 character hexadecimal string, got: {token}"
+                    )
                 headers["AuthToken"] = token
 
+        # Specific exceptions that should not be retried
+        fatal_exceptions = (
+            AuthTokenError,
+            InsufficientCreditsError,
+            ApidancePlatformError,
+        )
+
+        # Error storage
         last_error = None
+
         # Retry loop
         for attempt in range(1, self.max_retries + 1):
             try:
+                # Make the actual HTTP request
                 response = self.client.request(
                     method, url, headers=headers, timeout=10, **kwargs
                 )
@@ -204,36 +216,87 @@ class TwitterClient:
                     # Check if retry is needed
                     if not self._should_retry(response, attempt):
                         return response.json()
-                except (
-                    RateLimitError,
-                    InsufficientCreditsError,
-                    AuthenticationError,
-                ):  # These are explicit errors, raise immediately
+                except fatal_exceptions:
+                    # These are explicit errors that should not be retried
                     raise
+                except RateLimitError:
+                    # Rate limit errors are handled directly in _should_retry
+                    if attempt == self.max_retries:
+                        raise
+                    # If we're not at max retries, _should_retry will return True to continue
                 except Exception as e:
                     # Record other errors and continue retrying
                     last_error = e
+                    break
 
                 # Calculate delay for next retry
                 delay = self._calculate_retry_delay(attempt)
                 time.sleep(delay)
 
-            except httpx.ConnectTimeout:
-                raise TimeoutError("The handshake operation timed out") from None
+            except httpx.ConnectError as e:
+                if attempt == self.max_retries:
+                    raise TimeoutError(f"Connection error: {str(e)}") from e
+                # For network errors before max_retries, try again
+                last_error = e
+                delay = self._calculate_retry_delay(attempt)
+                time.sleep(delay)
+            except Exception as e:
+                # Catch any other unexpected errors
+                if attempt == self.max_retries:
+                    raise ApidancePlatformError(f"Unexpected error: {str(e)}") from e
+                last_error = e
+                delay = self._calculate_retry_delay(attempt)
+                time.sleep(delay)
 
-        # If all retries failed, raise the last error
+        # If all retries failed, raise the final error
         if last_error:
             raise last_error
-        return None
 
     def check_balance(self) -> int:
         """Check the remaining balance for the API key.
 
         Returns:
             int: Remaining balance
+
+        Note:
+            Returns 0 if connection fails or API returns error
         """
-        response = httpx.get(f"https://api.apidance.pro/key/{self.api_key}")
-        return response.text
+        try:
+            response = httpx.get(
+                f"https://api.apidance.pro/key/{self.api_key}", timeout=5.0, verify=True
+            )
+
+            # Try to parse as JSON
+            try:
+                data = response.json()
+                # If the response is a direct integer
+                if isinstance(data, int):
+                    return data
+                # If the response is an error object
+                if isinstance(data, dict) and data.get("code") == -1:
+                    return 0
+                # Any other JSON response, try to convert text to int
+                return int(response.text)
+            except (json.JSONDecodeError, ValueError):
+                # If not valid JSON but has text, try parsing as int
+                if response.text:
+                    try:
+                        return int(response.text)
+                    except ValueError:
+                        return 0
+                return 0
+
+        except (
+            httpx.ConnectError,
+            httpx.ConnectTimeout,
+            httpx.ReadTimeout,
+            httpx.HTTPError,
+            httpx.HTTPStatusError,
+        ) as e:
+            # Log the error if needed
+            print(f"Warning: Could not connect to API server: {e}")
+            # Return 0 as default balance when unable to check
+            return 0
 
     def search_timeline(
         self,
@@ -361,7 +424,7 @@ class TwitterClient:
 
     def get_list_latest_tweets(
         self,
-        list_id: str,
+        list_id: Union[int, str],
         count: int = 20,
         include_promoted_content: bool = False,
     ) -> List[Tweet]:
@@ -380,7 +443,7 @@ class TwitterClient:
 
         while True:
             variables = {
-                "listId": list_id,
+                "listId": str(list_id),
                 "count": count,
                 "includePromotedContent": include_promoted_content,
             }
@@ -419,9 +482,9 @@ class TwitterClient:
                 # Extract cursor for next page
                 elif (
                     content.get("__typename") == "TimelineTimelineCursor"
-                    and content["cursorType"] == "Bottom"
+                    and content.get("cursorType") == "Bottom"
                 ):
-                    cursor = content["value"]
+                    cursor = content.get("value")
 
                 # Handle tweets in modules
                 elif content.get("__typename") == "TimelineTimelineModule":
@@ -533,7 +596,7 @@ class TwitterClient:
 
     def get_user_tweets(
         self,
-        user_id: str,
+        user_id: Union[int, str],
         count: int = 20,
         include_pins: bool = True,
         include_promoted_content: bool = False,
@@ -560,7 +623,7 @@ class TwitterClient:
 
         while True:
             variables = {
-                "userId": user_id,
+                "userId": str(user_id),
                 "count": batch_size,
                 "cursor": cursor,
                 "includePromotedContent": include_promoted_content,
@@ -603,7 +666,7 @@ class TwitterClient:
 
         return all_tweets
 
-    def get_following(self, user_id: str) -> List[User]:
+    def get_following(self, user_id: Union[int, str]) -> List[User]:
         """Get a list of users that the specified user is following.
 
         Args:
@@ -614,7 +677,7 @@ class TwitterClient:
         """
 
         variables = {
-            "userId": user_id,
+            "userId": str(user_id),
             "includePromotedContent": False,
         }
 
@@ -654,7 +717,7 @@ class TwitterClient:
 
     def get_followers(
         self,
-        user_id: str,
+        user_id: Union[int, str],
         count: int = 20,
         include_promoted_content: bool = False,
     ) -> List[User]:
@@ -669,7 +732,7 @@ class TwitterClient:
             List of User objects containing detailed user information
         """
         variables = {
-            "userId": user_id,
+            "userId": str(user_id),
             "count": min(count, 20) if count > 0 else 20,
             "includePromotedContent": include_promoted_content,
         }
@@ -732,7 +795,7 @@ class TwitterClient:
 
     def get_followers_you_know(
         self,
-        user_id: str,
+        user_id: Union[int, str],
         count: int = 20,
         include_promoted_content: bool = False,
     ) -> List[User]:
@@ -747,7 +810,7 @@ class TwitterClient:
             List of User objects containing detailed user information
         """
         variables = {
-            "userId": user_id,
+            "userId": str(user_id),
             "count": min(count, 20) if count > 0 else 20,
             "includePromotedContent": include_promoted_content,
         }
@@ -808,10 +871,10 @@ class TwitterClient:
 
         return all_followers
 
-    def favorite_tweet(self, tweet_id: str) -> bool:
+    def favorite_tweet(self, tweet_id: Union[int, str]) -> bool:
 
         variables = {
-            "tweet_id": tweet_id,
+            "tweet_id": str(tweet_id),
         }
 
         response = self._make_request(
@@ -819,7 +882,6 @@ class TwitterClient:
             "/graphql/FavoriteTweet",
             json={"variables": variables},
         )
-
         if response == {"data": {"favorite_tweet": "Done"}}:
             print(f"Tweet: {tweet_id} favorited")
             return True
@@ -838,7 +900,7 @@ class TwitterClient:
 
         return False
 
-    def create_tweet(self, text: str, reply_to_tweet_id: str = None) -> str:
+    def create_tweet(self, text: str, reply_to_tweet_id: Union[int, str] = None) -> str:
         """Create a new tweet or reply to an existing tweet.
 
         Args:
@@ -858,7 +920,7 @@ class TwitterClient:
         # Add reply information if replying to a tweet
         if reply_to_tweet_id:
             variables["reply"] = {
-                "in_reply_to_tweet_id": reply_to_tweet_id,
+                "in_reply_to_tweet_id": str(reply_to_tweet_id),
                 "exclude_reply_user_ids": [],
             }
 
@@ -879,9 +941,68 @@ class TwitterClient:
             print("Failed to create tweet")
             return None
 
-    def tweet_result_by_rest_id(self, tweet_id: str):
+    def create_note_tweet(
+        self,
+        text: str,
+        use_richtext: bool = False,
+        reply_to_tweet_id: Union[int, str] = None,
+    ) -> str:
+        """Create a new note tweet. (Long text tweet, only available in Premium+)
+
+        Args:
+            text: The text content of the tweet (supports markdown if use_richtext=True)
+            use_richtext: Whether to parse markdown formatting for rich text
+            reply_to_tweet_id: Optional tweet ID to reply to
+
+        Returns:
+            str: The ID of the created tweet
+
+        Raises:
+            PremiumRequiredError: If the user doesn't have Premium+ plan
+        """
+        richtext_options = {}
+
         variables = {
-            "tweetId": tweet_id,
+            "tweet_text": text,
+            "dark_request": False,
+            "media": {"media_entities": [], "possibly_sensitive": False},
+            "semantic_annotation_ids": [],
+            "includePromotedContent": False,
+        }
+
+        if use_richtext:
+            text, richtext_tags = parse_markdown_to_richtext(text)
+            if richtext_tags:
+                richtext_options["richtext_tags"] = richtext_tags
+                variables["richtext_options"] = richtext_options
+
+        # Add reply information if replying to a tweet
+        if reply_to_tweet_id:
+            variables["reply"] = {
+                "in_reply_to_tweet_id": str(reply_to_tweet_id),
+                "exclude_reply_user_ids": [],
+            }
+
+        response = self._make_request(
+            "POST",
+            "/graphql/CreateNoteTweet",
+            json={"variables": variables},
+        )
+
+        # Extract the tweet ID from the response
+        try:
+            tweet_id = response["data"]["notetweet_create"]["tweet_results"]["result"][
+                "rest_id"
+            ]
+            print(f"Note tweet created successfully with ID: {tweet_id}")
+            return tweet_id
+        except (KeyError, TypeError):
+            print("Failed to create note tweet")
+            return None
+
+    def tweet_result_by_rest_id(self, tweet_id: Union[int, str]):
+        variables = {
+            "tweetId": str(tweet_id),
             "withHighlightedLabel": True,
             "withTweetQuoteCount": True,
             "includePromotedContent": True,
@@ -896,6 +1017,7 @@ class TwitterClient:
             params={"variables": variables},
         )
 
+        # {'data': {'tweetResult': {}}}
         # response maybe []
 
         return Tweet.from_api_response(response["data"])
